using System;
using System.Collections;
using System.Threading;
using Intel.Utilities;
using Intel.UPNP;

namespace TemperatureControlPoint
{
    /// <summary>
    /// Transparent ClientSide UPnP Service
    /// </summary>
    public class CpLedsService
    {
        private Hashtable UnspecifiedTable = Hashtable.Synchronized(new Hashtable());
        internal UPnPService _S;

        public UPnPService GetUPnPService()
        {
            return (_S);
        }
        public static string SERVICE_NAME = "urn:schemas-upnp-org:service:LedsService:";
        public double VERSION
        {
            get
            {
                return (double.Parse(_S.Version));
            }
        }

        public delegate void StateVariableModifiedHandler_StatusBlueLed(CpLedsService sender, System.Boolean NewValue);
        private WeakEvent OnStateVariable_StatusBlueLed_Event = new WeakEvent();
        public event StateVariableModifiedHandler_StatusBlueLed OnStateVariable_StatusBlueLed
        {
            add { OnStateVariable_StatusBlueLed_Event.Register(value); }
            remove { OnStateVariable_StatusBlueLed_Event.UnRegister(value); }
        }
        protected void StatusBlueLed_ModifiedSink(UPnPStateVariable Var, object NewValue)
        {
            OnStateVariable_StatusBlueLed_Event.Fire(this, StatusBlueLed);
        }
        public delegate void StateVariableModifiedHandler_StatusRedLed(CpLedsService sender, System.Boolean NewValue);
        private WeakEvent OnStateVariable_StatusRedLed_Event = new WeakEvent();
        public event StateVariableModifiedHandler_StatusRedLed OnStateVariable_StatusRedLed
        {
            add { OnStateVariable_StatusRedLed_Event.Register(value); }
            remove { OnStateVariable_StatusRedLed_Event.UnRegister(value); }
        }
        protected void StatusRedLed_ModifiedSink(UPnPStateVariable Var, object NewValue)
        {
            OnStateVariable_StatusRedLed_Event.Fire(this, StatusRedLed);
        }
        public delegate void StateVariableModifiedHandler_StatusWhiteLed(CpLedsService sender, System.Boolean NewValue);
        private WeakEvent OnStateVariable_StatusWhiteLed_Event = new WeakEvent();
        public event StateVariableModifiedHandler_StatusWhiteLed OnStateVariable_StatusWhiteLed
        {
            add { OnStateVariable_StatusWhiteLed_Event.Register(value); }
            remove { OnStateVariable_StatusWhiteLed_Event.UnRegister(value); }
        }
        protected void StatusWhiteLed_ModifiedSink(UPnPStateVariable Var, object NewValue)
        {
            OnStateVariable_StatusWhiteLed_Event.Fire(this, StatusWhiteLed);
        }
        public delegate void SubscribeHandler(CpLedsService sender, bool Success);
        public event SubscribeHandler OnSubscribe;
        public delegate void Delegate_OnResult_GetBlueLedStatus(CpLedsService sender, System.Boolean ReturnValue, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_GetBlueLedStatus_Event = new WeakEvent();
        public event Delegate_OnResult_GetBlueLedStatus OnResult_GetBlueLedStatus
        {
            add { OnResult_GetBlueLedStatus_Event.Register(value); }
            remove { OnResult_GetBlueLedStatus_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_GetRedLedStatus(CpLedsService sender, System.Boolean ReturnValue, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_GetRedLedStatus_Event = new WeakEvent();
        public event Delegate_OnResult_GetRedLedStatus OnResult_GetRedLedStatus
        {
            add { OnResult_GetRedLedStatus_Event.Register(value); }
            remove { OnResult_GetRedLedStatus_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_GetWhiteLedStatus(CpLedsService sender, System.Boolean ReturnValue, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_GetWhiteLedStatus_Event = new WeakEvent();
        public event Delegate_OnResult_GetWhiteLedStatus OnResult_GetWhiteLedStatus
        {
            add { OnResult_GetWhiteLedStatus_Event.Register(value); }
            remove { OnResult_GetWhiteLedStatus_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OffBlueLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OffBlueLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OffBlueLeds OnResult_OffBlueLeds
        {
            add { OnResult_OffBlueLeds_Event.Register(value); }
            remove { OnResult_OffBlueLeds_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OffRedLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OffRedLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OffRedLeds OnResult_OffRedLeds
        {
            add { OnResult_OffRedLeds_Event.Register(value); }
            remove { OnResult_OffRedLeds_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OffWhiteLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OffWhiteLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OffWhiteLeds OnResult_OffWhiteLeds
        {
            add { OnResult_OffWhiteLeds_Event.Register(value); }
            remove { OnResult_OffWhiteLeds_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OnBlueLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OnBlueLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OnBlueLeds OnResult_OnBlueLeds
        {
            add { OnResult_OnBlueLeds_Event.Register(value); }
            remove { OnResult_OnBlueLeds_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OnRedLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OnRedLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OnRedLeds OnResult_OnRedLeds
        {
            add { OnResult_OnRedLeds_Event.Register(value); }
            remove { OnResult_OnRedLeds_Event.UnRegister(value); }
        }
        public delegate void Delegate_OnResult_OnWhiteLeds(CpLedsService sender, UPnPInvokeException e, object _Tag);
        private WeakEvent OnResult_OnWhiteLeds_Event = new WeakEvent();
        public event Delegate_OnResult_OnWhiteLeds OnResult_OnWhiteLeds
        {
            add { OnResult_OnWhiteLeds_Event.Register(value); }
            remove { OnResult_OnWhiteLeds_Event.UnRegister(value); }
        }

        public CpLedsService(UPnPService s)
        {
            _S = s;
            _S.OnSubscribe += new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
            if (HasStateVariable_StatusBlueLed) _S.GetStateVariableObject("StatusBlueLed").OnModified += new UPnPStateVariable.ModifiedHandler(StatusBlueLed_ModifiedSink);
            if (HasStateVariable_StatusRedLed) _S.GetStateVariableObject("StatusRedLed").OnModified += new UPnPStateVariable.ModifiedHandler(StatusRedLed_ModifiedSink);
            if (HasStateVariable_StatusWhiteLed) _S.GetStateVariableObject("StatusWhiteLed").OnModified += new UPnPStateVariable.ModifiedHandler(StatusWhiteLed_ModifiedSink);
        }
        public void Dispose()
        {
            _S.OnSubscribe -= new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
            OnSubscribe = null;
            if (HasStateVariable_StatusBlueLed) _S.GetStateVariableObject("StatusBlueLed").OnModified -= new UPnPStateVariable.ModifiedHandler(StatusBlueLed_ModifiedSink);
            if (HasStateVariable_StatusRedLed) _S.GetStateVariableObject("StatusRedLed").OnModified -= new UPnPStateVariable.ModifiedHandler(StatusRedLed_ModifiedSink);
            if (HasStateVariable_StatusWhiteLed) _S.GetStateVariableObject("StatusWhiteLed").OnModified -= new UPnPStateVariable.ModifiedHandler(StatusWhiteLed_ModifiedSink);
        }
        public void _subscribe(int Timeout)
        {
            _S.Subscribe(Timeout, null);
        }
        protected void _subscribe_sink(UPnPService sender, bool OK)
        {
            if (OnSubscribe != null)
            {
                OnSubscribe(this, OK);
            }
        }
        public void SetUnspecifiedValue(string EnumType, string val)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            UnspecifiedTable[hash] = val;
        }
        public string GetUnspecifiedValue(string EnumType)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            if (UnspecifiedTable.ContainsKey(hash) == false)
            {
                return ("");
            }
            string RetVal = (string)UnspecifiedTable[hash];
            return (RetVal);
        }
        public System.Boolean StatusBlueLed
        {
            get
            {
                return ((System.Boolean)_S.GetStateVariable("StatusBlueLed"));
            }
        }
        public System.Boolean StatusRedLed
        {
            get
            {
                return ((System.Boolean)_S.GetStateVariable("StatusRedLed"));
            }
        }
        public System.Boolean StatusWhiteLed
        {
            get
            {
                return ((System.Boolean)_S.GetStateVariable("StatusWhiteLed"));
            }
        }
        public bool HasStateVariable_StatusBlueLed
        {
            get
            {
                if (_S.GetStateVariableObject("StatusBlueLed") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasStateVariable_StatusRedLed
        {
            get
            {
                if (_S.GetStateVariableObject("StatusRedLed") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasStateVariable_StatusWhiteLed
        {
            get
            {
                if (_S.GetStateVariableObject("StatusWhiteLed") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_GetBlueLedStatus
        {
            get
            {
                if (_S.GetAction("GetBlueLedStatus") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_GetRedLedStatus
        {
            get
            {
                if (_S.GetAction("GetRedLedStatus") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_GetWhiteLedStatus
        {
            get
            {
                if (_S.GetAction("GetWhiteLedStatus") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OffBlueLeds
        {
            get
            {
                if (_S.GetAction("OffBlueLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OffRedLeds
        {
            get
            {
                if (_S.GetAction("OffRedLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OffWhiteLeds
        {
            get
            {
                if (_S.GetAction("OffWhiteLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OnBlueLeds
        {
            get
            {
                if (_S.GetAction("OnBlueLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OnRedLeds
        {
            get
            {
                if (_S.GetAction("OnRedLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public bool HasAction_OnWhiteLeds
        {
            get
            {
                if (_S.GetAction("OnWhiteLeds") == null)
                {
                    return (false);
                }
                else
                {
                    return (true);
                }
            }
        }
        public System.Boolean Sync_GetBlueLedStatus()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            object RetVal = null;
            RetVal = _S.InvokeSync("GetBlueLedStatus", args);
            return ((System.Boolean)RetVal);
        }
        public void GetBlueLedStatus()
        {
            GetBlueLedStatus(null, null);
        }
        public void GetBlueLedStatus(object _Tag, Delegate_OnResult_GetBlueLedStatus _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("GetBlueLedStatus", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_GetBlueLedStatus), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_GetBlueLedStatus));
        }
        private void Sink_GetBlueLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetBlueLedStatus)StateInfo[1])(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
            else
            {
                OnResult_GetBlueLedStatus_Event.Fire(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
        }
        private void Error_Sink_GetBlueLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetBlueLedStatus)StateInfo[1])(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
            else
            {
                OnResult_GetBlueLedStatus_Event.Fire(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
        }
        public System.Boolean Sync_GetRedLedStatus()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            object RetVal = null;
            RetVal = _S.InvokeSync("GetRedLedStatus", args);
            return ((System.Boolean)RetVal);
        }
        public void GetRedLedStatus()
        {
            GetRedLedStatus(null, null);
        }
        public void GetRedLedStatus(object _Tag, Delegate_OnResult_GetRedLedStatus _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("GetRedLedStatus", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_GetRedLedStatus), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_GetRedLedStatus));
        }
        private void Sink_GetRedLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetRedLedStatus)StateInfo[1])(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
            else
            {
                OnResult_GetRedLedStatus_Event.Fire(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
        }
        private void Error_Sink_GetRedLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetRedLedStatus)StateInfo[1])(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
            else
            {
                OnResult_GetRedLedStatus_Event.Fire(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
        }
        public System.Boolean Sync_GetWhiteLedStatus()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            object RetVal = null;
            RetVal = _S.InvokeSync("GetWhiteLedStatus", args);
            return ((System.Boolean)RetVal);
        }
        public void GetWhiteLedStatus()
        {
            GetWhiteLedStatus(null, null);
        }
        public void GetWhiteLedStatus(object _Tag, Delegate_OnResult_GetWhiteLedStatus _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("GetWhiteLedStatus", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_GetWhiteLedStatus), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_GetWhiteLedStatus));
        }
        private void Sink_GetWhiteLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetWhiteLedStatus)StateInfo[1])(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
            else
            {
                OnResult_GetWhiteLedStatus_Event.Fire(this, (System.Boolean)RetVal, null, StateInfo[0]);
            }
        }
        private void Error_Sink_GetWhiteLedStatus(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_GetWhiteLedStatus)StateInfo[1])(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
            else
            {
                OnResult_GetWhiteLedStatus_Event.Fire(this, (System.Boolean)UPnPService.CreateObjectInstance(typeof(System.Boolean), null), e, StateInfo[0]);
            }
        }
        public void Sync_OffBlueLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OffBlueLeds", args);
            return;
        }
        public void OffBlueLeds()
        {
            OffBlueLeds(null, null);
        }
        public void OffBlueLeds(object _Tag, Delegate_OnResult_OffBlueLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OffBlueLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OffBlueLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OffBlueLeds));
        }
        private void Sink_OffBlueLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffBlueLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OffBlueLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OffBlueLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffBlueLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OffBlueLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_OffRedLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OffRedLeds", args);
            return;
        }
        public void OffRedLeds()
        {
            OffRedLeds(null, null);
        }
        public void OffRedLeds(object _Tag, Delegate_OnResult_OffRedLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OffRedLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OffRedLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OffRedLeds));
        }
        private void Sink_OffRedLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffRedLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OffRedLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OffRedLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffRedLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OffRedLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_OffWhiteLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OffWhiteLeds", args);
            return;
        }
        public void OffWhiteLeds()
        {
            OffWhiteLeds(null, null);
        }
        public void OffWhiteLeds(object _Tag, Delegate_OnResult_OffWhiteLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OffWhiteLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OffWhiteLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OffWhiteLeds));
        }
        private void Sink_OffWhiteLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffWhiteLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OffWhiteLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OffWhiteLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OffWhiteLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OffWhiteLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_OnBlueLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OnBlueLeds", args);
            return;
        }
        public void OnBlueLeds()
        {
            OnBlueLeds(null, null);
        }
        public void OnBlueLeds(object _Tag, Delegate_OnResult_OnBlueLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OnBlueLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OnBlueLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OnBlueLeds));
        }
        private void Sink_OnBlueLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnBlueLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OnBlueLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OnBlueLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnBlueLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OnBlueLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_OnRedLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OnRedLeds", args);
            return;
        }
        public void OnRedLeds()
        {
            OnRedLeds(null, null);
        }
        public void OnRedLeds(object _Tag, Delegate_OnResult_OnRedLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OnRedLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OnRedLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OnRedLeds));
        }
        private void Sink_OnRedLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnRedLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OnRedLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OnRedLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnRedLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OnRedLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_OnWhiteLeds()
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("OnWhiteLeds", args);
            return;
        }
        public void OnWhiteLeds()
        {
            OnWhiteLeds(null, null);
        }
        public void OnWhiteLeds(object _Tag, Delegate_OnResult_OnWhiteLeds _Callback)
        {
            UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeAsync("OnWhiteLeds", args, new object[2] { _Tag, _Callback }, new UPnPService.UPnPServiceInvokeHandler(Sink_OnWhiteLeds), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_OnWhiteLeds));
        }
        private void Sink_OnWhiteLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnWhiteLeds)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_OnWhiteLeds_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_OnWhiteLeds(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if (StateInfo[1] != null)
            {
                ((Delegate_OnResult_OnWhiteLeds)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_OnWhiteLeds_Event.Fire(this, e, StateInfo[0]);
            }
        }
    }
}