using System;
using System.Collections;
using System.Threading;
using Intel.Utilities;
using Intel.UPNP;

namespace TemperatureControlPoint
{
    /// <summary>
    /// Transparent ClientSide UPnP Service
    /// </summary>
    public class CpMusicService
    {
       private Hashtable UnspecifiedTable = Hashtable.Synchronized(new Hashtable());
       internal UPnPService _S;

       public UPnPService GetUPnPService()
       {
            return(_S);
       }
       public static string SERVICE_NAME = "urn:schemas-upnp-org:service:MusicService:";
       public double VERSION
       {
           get
           {
               return(double.Parse(_S.Version));
           }
       }

       public delegate void SubscribeHandler(CpMusicService sender, bool Success);
       public event SubscribeHandler OnSubscribe;
       public delegate void Delegate_OnResult_assignAddress(CpMusicService sender, System.String address, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_assignAddress_Event = new WeakEvent();
       public event Delegate_OnResult_assignAddress OnResult_assignAddress
       {
			add{OnResult_assignAddress_Event.Register(value);}
			remove{OnResult_assignAddress_Event.UnRegister(value);}
       }
       public delegate void Delegate_OnResult_play(CpMusicService sender, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_play_Event = new WeakEvent();
       public event Delegate_OnResult_play OnResult_play
       {
			add{OnResult_play_Event.Register(value);}
			remove{OnResult_play_Event.UnRegister(value);}
       }
       public delegate void Delegate_OnResult_stop(CpMusicService sender, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_stop_Event = new WeakEvent();
       public event Delegate_OnResult_stop OnResult_stop
       {
			add{OnResult_stop_Event.Register(value);}
			remove{OnResult_stop_Event.UnRegister(value);}
       }

        public CpMusicService(UPnPService s)
        {
            _S = s;
            _S.OnSubscribe += new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
        }
        public void Dispose()
        {
            _S.OnSubscribe -= new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
            OnSubscribe = null;
        }
        public void _subscribe(int Timeout)
        {
            _S.Subscribe(Timeout, null);
        }
        protected void _subscribe_sink(UPnPService sender, bool OK)
        {
            if(OnSubscribe!=null)
            {
                OnSubscribe(this, OK);
            }
        }
        public void SetUnspecifiedValue(string EnumType, string val)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            UnspecifiedTable[hash] = val;
        }
        public string GetUnspecifiedValue(string EnumType)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            if(UnspecifiedTable.ContainsKey(hash)==false)
            {
               return("");
            }
            string RetVal = (string)UnspecifiedTable[hash];
            return(RetVal);
        }
        public System.String Address
        {
            get
            {
               return((System.String)_S.GetStateVariable("Address"));
            }
        }
        public bool HasStateVariable_Address
        {
            get
            {
               if(_S.GetStateVariableObject("Address")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_assignAddress
        {
            get
            {
               if(_S.GetAction("assignAddress")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_play
        {
            get
            {
               if(_S.GetAction("play")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_stop
        {
            get
            {
               if(_S.GetAction("stop")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public void Sync_assignAddress(System.String address)
        {
           UPnPArgument[] args = new UPnPArgument[1];
           args[0] = new UPnPArgument("address", address);
            _S.InvokeSync("assignAddress", args);
            address = (System.String) args[0].DataValue;
            return;
        }
        public void assignAddress(System.String address)
        {
            assignAddress(address, null, null);
        }
        public void assignAddress(System.String address, object _Tag, Delegate_OnResult_assignAddress _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[1];
           args[0] = new UPnPArgument("address", address);
           _S.InvokeAsync("assignAddress", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_assignAddress), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_assignAddress));
        }
        private void Sink_assignAddress(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_assignAddress)StateInfo[1])(this, (System.String )Args[0].DataValue, null, StateInfo[0]);
            }
            else
            {
                OnResult_assignAddress_Event.Fire(this, (System.String )Args[0].DataValue, null, StateInfo[0]);
            }
        }
        private void Error_Sink_assignAddress(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_assignAddress)StateInfo[1])(this, (System.String )Args[0].DataValue, e, StateInfo[0]);
            }
            else
            {
                OnResult_assignAddress_Event.Fire(this, (System.String )Args[0].DataValue, e, StateInfo[0]);
            }
        }
        public void Sync_play()
        {
           UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("play", args);
            return;
        }
        public void play()
        {
            play(null, null);
        }
        public void play(object _Tag, Delegate_OnResult_play _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[0];
           _S.InvokeAsync("play", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_play), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_play));
        }
        private void Sink_play(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_play)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_play_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_play(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_play)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_play_Event.Fire(this, e, StateInfo[0]);
            }
        }
        public void Sync_stop()
        {
           UPnPArgument[] args = new UPnPArgument[0];
            _S.InvokeSync("stop", args);
            return;
        }
        public void stop()
        {
            stop(null, null);
        }
        public void stop(object _Tag, Delegate_OnResult_stop _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[0];
           _S.InvokeAsync("stop", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_stop), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_stop));
        }
        private void Sink_stop(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_stop)StateInfo[1])(this, null, StateInfo[0]);
            }
            else
            {
                OnResult_stop_Event.Fire(this, null, StateInfo[0]);
            }
        }
        private void Error_Sink_stop(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_stop)StateInfo[1])(this, e, StateInfo[0]);
            }
            else
            {
                OnResult_stop_Event.Fire(this, e, StateInfo[0]);
            }
        }
    }
}