using System;
using System.Collections;
using System.Threading;
using Intel.Utilities;
using Intel.UPNP;

namespace Intel.DeviceBuilder
{
    /// <summary>
    /// Transparent ClientSide UPnP Service
    /// </summary>
    public class CpConfigService
    {
       private Hashtable UnspecifiedTable = Hashtable.Synchronized(new Hashtable());
       internal UPnPService _S;

       public UPnPService GetUPnPService()
       {
            return(_S);
       }
       public static string SERVICE_NAME = "urn:schemas-upnp-org:service:ConfigService:";
       public double VERSION
       {
           get
           {
               return(double.Parse(_S.Version));
           }
       }

       public delegate void SubscribeHandler(CpConfigService sender, bool Success);
       public event SubscribeHandler OnSubscribe;
       public delegate void Delegate_OnResult_configure(CpConfigService sender, System.String config, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_configure_Event = new WeakEvent();
       public event Delegate_OnResult_configure OnResult_configure
       {
			add{OnResult_configure_Event.Register(value);}
			remove{OnResult_configure_Event.UnRegister(value);}
       }

        public CpConfigService(UPnPService s)
        {
            _S = s;
            _S.OnSubscribe += new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
        }
        public void Dispose()
        {
            _S.OnSubscribe -= new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
            OnSubscribe = null;
        }
        public void _subscribe(int Timeout)
        {
            _S.Subscribe(Timeout, null);
        }
        protected void _subscribe_sink(UPnPService sender, bool OK)
        {
            if(OnSubscribe!=null)
            {
                OnSubscribe(this, OK);
            }
        }
        public void SetUnspecifiedValue(string EnumType, string val)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            UnspecifiedTable[hash] = val;
        }
        public string GetUnspecifiedValue(string EnumType)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            if(UnspecifiedTable.ContainsKey(hash)==false)
            {
               return("");
            }
            string RetVal = (string)UnspecifiedTable[hash];
            return(RetVal);
        }
        public System.String Config
        {
            get
            {
               return((System.String)_S.GetStateVariable("Config"));
            }
        }
        public bool HasStateVariable_Config
        {
            get
            {
               if(_S.GetStateVariableObject("Config")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_configure
        {
            get
            {
               if(_S.GetAction("configure")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public void Sync_configure(System.String config)
        {
           UPnPArgument[] args = new UPnPArgument[1];
           args[0] = new UPnPArgument("config", config);
            _S.InvokeSync("configure", args);
            config = (System.String) args[0].DataValue;
            return;
        }
        public void configure(System.String config)
        {
            configure(config, null, null);
        }
        public void configure(System.String config, object _Tag, Delegate_OnResult_configure _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[1];
           args[0] = new UPnPArgument("config", config);
           _S.InvokeAsync("configure", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_configure), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_configure));
        }
        private void Sink_configure(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_configure)StateInfo[1])(this, (System.String )Args[0].DataValue, null, StateInfo[0]);
            }
            else
            {
                OnResult_configure_Event.Fire(this, (System.String )Args[0].DataValue, null, StateInfo[0]);
            }
        }
        private void Error_Sink_configure(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_configure)StateInfo[1])(this, (System.String )Args[0].DataValue, e, StateInfo[0]);
            }
            else
            {
                OnResult_configure_Event.Fire(this, (System.String )Args[0].DataValue, e, StateInfo[0]);
            }
        }
    }
}