using System;
using System.Collections;
using System.Threading;
using Intel.Utilities;
using Intel.UPNP;

namespace TemperatureControlPoint
{
    /// <summary>
    /// Transparent ClientSide UPnP Service
    /// </summary>
    public class CpTemperatureService
    {
       private Hashtable UnspecifiedTable = Hashtable.Synchronized(new Hashtable());
       internal UPnPService _S;

       public UPnPService GetUPnPService()
       {
            return(_S);
       }
       public static string SERVICE_NAME = "urn:schemas-upnp-org:service:TemperatureService:";
       public double VERSION
       {
           get
           {
               return(double.Parse(_S.Version));
           }
       }

       public delegate void SubscribeHandler(CpTemperatureService sender, bool Success);
       public event SubscribeHandler OnSubscribe;
       public delegate void Delegate_OnResult_getTemperatureAmbiante(CpTemperatureService sender, System.Int32 ReturnValue, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_getTemperatureAmbiante_Event = new WeakEvent();
       public event Delegate_OnResult_getTemperatureAmbiante OnResult_getTemperatureAmbiante
       {
			add{OnResult_getTemperatureAmbiante_Event.Register(value);}
			remove{OnResult_getTemperatureAmbiante_Event.UnRegister(value);}
       }
       public delegate void Delegate_OnResult_getTemperatureEau(CpTemperatureService sender, System.Int32 ReturnValue, UPnPInvokeException e, object _Tag);
       private WeakEvent OnResult_getTemperatureEau_Event = new WeakEvent();
       public event Delegate_OnResult_getTemperatureEau OnResult_getTemperatureEau
       {
			add{OnResult_getTemperatureEau_Event.Register(value);}
			remove{OnResult_getTemperatureEau_Event.UnRegister(value);}
       }

        public CpTemperatureService(UPnPService s)
        {
            _S = s;
            _S.OnSubscribe += new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
        }
        public void Dispose()
        {
            _S.OnSubscribe -= new UPnPService.UPnPEventSubscribeHandler(_subscribe_sink);
            OnSubscribe = null;
        }
        public void _subscribe(int Timeout)
        {
            _S.Subscribe(Timeout, null);
        }
        protected void _subscribe_sink(UPnPService sender, bool OK)
        {
            if(OnSubscribe!=null)
            {
                OnSubscribe(this, OK);
            }
        }
        public void SetUnspecifiedValue(string EnumType, string val)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            UnspecifiedTable[hash] = val;
        }
        public string GetUnspecifiedValue(string EnumType)
        {
            string hash = Thread.CurrentThread.GetHashCode().ToString() + ":" + EnumType;
            if(UnspecifiedTable.ContainsKey(hash)==false)
            {
               return("");
            }
            string RetVal = (string)UnspecifiedTable[hash];
            return(RetVal);
        }
        public System.Int32 temperatureAmbiante
        {
            get
            {
               return((System.Int32)_S.GetStateVariable("temperatureAmbiante"));
            }
        }
        public System.Int32 temperatureEau
        {
            get
            {
               return((System.Int32)_S.GetStateVariable("temperatureEau"));
            }
        }
        public bool HasStateVariable_temperatureAmbiante
        {
            get
            {
               if(_S.GetStateVariableObject("temperatureAmbiante")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasStateVariable_temperatureEau
        {
            get
            {
               if(_S.GetStateVariableObject("temperatureEau")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_getTemperatureAmbiante
        {
            get
            {
               if(_S.GetAction("getTemperatureAmbiante")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public bool HasAction_getTemperatureEau
        {
            get
            {
               if(_S.GetAction("getTemperatureEau")==null)
               {
                   return(false);
               }
               else
               {
                   return(true);
               }
            }
        }
        public System.Int32 Sync_getTemperatureAmbiante()
        {
           UPnPArgument[] args = new UPnPArgument[0];
           object RetVal = null;
           RetVal = _S.InvokeSync("getTemperatureAmbiante", args);
           return((System.Int32)RetVal);
        }
        public void getTemperatureAmbiante()
        {
            getTemperatureAmbiante(null, null);
        }
        public void getTemperatureAmbiante(object _Tag, Delegate_OnResult_getTemperatureAmbiante _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[0];
           _S.InvokeAsync("getTemperatureAmbiante", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_getTemperatureAmbiante), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_getTemperatureAmbiante));
        }
        private void Sink_getTemperatureAmbiante(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_getTemperatureAmbiante)StateInfo[1])(this, (System.Int32)RetVal, null, StateInfo[0]);
            }
            else
            {
                OnResult_getTemperatureAmbiante_Event.Fire(this, (System.Int32)RetVal, null, StateInfo[0]);
            }
        }
        private void Error_Sink_getTemperatureAmbiante(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_getTemperatureAmbiante)StateInfo[1])(this, (System.Int32)UPnPService.CreateObjectInstance(typeof(System.Int32),null), e, StateInfo[0]);
            }
            else
            {
                OnResult_getTemperatureAmbiante_Event.Fire(this, (System.Int32)UPnPService.CreateObjectInstance(typeof(System.Int32),null), e, StateInfo[0]);
            }
        }
        public System.Int32 Sync_getTemperatureEau()
        {
           UPnPArgument[] args = new UPnPArgument[0];
           object RetVal = null;
           RetVal = _S.InvokeSync("getTemperatureEau", args);
           return((System.Int32)RetVal);
        }
        public void getTemperatureEau()
        {
            getTemperatureEau(null, null);
        }
        public void getTemperatureEau(object _Tag, Delegate_OnResult_getTemperatureEau _Callback)
        {
           UPnPArgument[] args = new UPnPArgument[0];
           _S.InvokeAsync("getTemperatureEau", args, new object[2]{_Tag,_Callback},new UPnPService.UPnPServiceInvokeHandler(Sink_getTemperatureEau), new UPnPService.UPnPServiceInvokeErrorHandler(Error_Sink_getTemperatureEau));
        }
        private void Sink_getTemperatureEau(UPnPService sender, string MethodName, UPnPArgument[] Args, object RetVal, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_getTemperatureEau)StateInfo[1])(this, (System.Int32)RetVal, null, StateInfo[0]);
            }
            else
            {
                OnResult_getTemperatureEau_Event.Fire(this, (System.Int32)RetVal, null, StateInfo[0]);
            }
        }
        private void Error_Sink_getTemperatureEau(UPnPService sender, string MethodName, UPnPArgument[] Args, UPnPInvokeException e, object _Tag)
        {
            object[] StateInfo = (object[])_Tag;
            if(StateInfo[1]!=null)
            {
                ((Delegate_OnResult_getTemperatureEau)StateInfo[1])(this, (System.Int32)UPnPService.CreateObjectInstance(typeof(System.Int32),null), e, StateInfo[0]);
            }
            else
            {
                OnResult_getTemperatureEau_Event.Fire(this, (System.Int32)UPnPService.CreateObjectInstance(typeof(System.Int32),null), e, StateInfo[0]);
            }
        }
    }
}