/*   
Copyright 2006 - 2010 Intel Corporation

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#ifndef __UPnPMicrostack__
#define __UPnPMicrostack__


#include "ILibAsyncSocket.h"

/*! \file UPnPMicroStack.h 
	\brief MicroStack APIs for Device Implementation
*/

/*! \defgroup MicroStack MicroStack Module
	\{
*/

struct UPnPDataObject;
struct packetheader;

typedef void* UPnPMicroStackToken;
typedef void* UPnPSessionToken;



/* Complex Type Parsers */


/* Complex Type Serializers */



/* UPnP Stack Management */
UPnPMicroStackToken UPnPCreateMicroStack(void *Chain, const char* FriendlyName, const char* UDN, const char* SerialNumber, const int NotifyCycleSeconds, const unsigned short PortNum);


void UPnPIPAddressListChanged(UPnPMicroStackToken MicroStackToken);
int UPnPGetLocalPortNumber(UPnPSessionToken token);
int   UPnPGetLocalInterfaceToHost(const UPnPSessionToken UPnPToken);
void* UPnPGetWebServerToken(const UPnPMicroStackToken MicroStackToken);
void UPnPSetTag(const UPnPMicroStackToken token, void *UserToken);
void *UPnPGetTag(const UPnPMicroStackToken token);
UPnPMicroStackToken UPnPGetMicroStackTokenFromSessionToken(const UPnPSessionToken token);

typedef void(*UPnP_ActionHandler_LedsService_GetBlueLedStatus) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_GetRedLedStatus) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_GetWhiteLedStatus) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OffBlueLeds) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OffRedLeds) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OffWhiteLeds) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OnBlueLeds) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OnRedLeds) (void* upnptoken);
typedef void(*UPnP_ActionHandler_LedsService_OnWhiteLeds) (void* upnptoken);
/* UPnP Set Function Pointers Methods */
extern void (*UPnPFP_PresentationPage) (void* upnptoken,struct packetheader *packet);
extern UPnP_ActionHandler_LedsService_GetBlueLedStatus UPnPFP_LedsService_GetBlueLedStatus;
extern UPnP_ActionHandler_LedsService_GetRedLedStatus UPnPFP_LedsService_GetRedLedStatus;
extern UPnP_ActionHandler_LedsService_GetWhiteLedStatus UPnPFP_LedsService_GetWhiteLedStatus;
extern UPnP_ActionHandler_LedsService_OffBlueLeds UPnPFP_LedsService_OffBlueLeds;
extern UPnP_ActionHandler_LedsService_OffRedLeds UPnPFP_LedsService_OffRedLeds;
extern UPnP_ActionHandler_LedsService_OffWhiteLeds UPnPFP_LedsService_OffWhiteLeds;
extern UPnP_ActionHandler_LedsService_OnBlueLeds UPnPFP_LedsService_OnBlueLeds;
extern UPnP_ActionHandler_LedsService_OnRedLeds UPnPFP_LedsService_OnRedLeds;
extern UPnP_ActionHandler_LedsService_OnWhiteLeds UPnPFP_LedsService_OnWhiteLeds;


void UPnPSetDisconnectFlag(UPnPSessionToken token,void *flag);

/* Invocation Response Methods */
void UPnPResponse_Error(const UPnPSessionToken UPnPToken, const int ErrorCode, const char* ErrorMsg);
void UPnPResponseGeneric(const UPnPSessionToken UPnPToken,const char* ServiceURI,const char* MethodName,const char* Params);
void UPnPResponse_LedsService_GetBlueLedStatus(const UPnPSessionToken UPnPToken, const int statusBlueLed);
void UPnPResponse_LedsService_GetRedLedStatus(const UPnPSessionToken UPnPToken, const int statusRedLed);
void UPnPResponse_LedsService_GetWhiteLedStatus(const UPnPSessionToken UPnPToken, const int statusWhiteLed);
void UPnPResponse_LedsService_OffBlueLeds(const UPnPSessionToken UPnPToken);
void UPnPResponse_LedsService_OffRedLeds(const UPnPSessionToken UPnPToken);
void UPnPResponse_LedsService_OffWhiteLeds(const UPnPSessionToken UPnPToken);
void UPnPResponse_LedsService_OnBlueLeds(const UPnPSessionToken UPnPToken);
void UPnPResponse_LedsService_OnRedLeds(const UPnPSessionToken UPnPToken);
void UPnPResponse_LedsService_OnWhiteLeds(const UPnPSessionToken UPnPToken);

/* State Variable Eventing Methods */
void UPnPSetState_LedsService_StatusRedLed(UPnPMicroStackToken microstack,int val);
void UPnPSetState_LedsService_StatusBlueLed(UPnPMicroStackToken microstack,int val);
void UPnPSetState_LedsService_StatusWhiteLed(UPnPMicroStackToken microstack,int val);






/*! \} */
#endif
